/*
 * GPIO sysfs user space application
 *
 * Copyright (C) 2012 Texas Instruments Incorporated - http://www.ti.com/
 * AnilKumar Ch <anilkumar@ti.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation version 2.
 *
 * This program is distributed "as is" WITHOUT ANY WARRANTY of any
 * kind, whether express or implied; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <getopt.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <poll.h>

#include "gpio-sysfs.h"

#define SYSFS_GPIO_DIR		"/sys/class/gpio"
#define POLL_TIMEOUT		(20 * 1000) /* 20 seconds */

#define MAX_BUF_SIZE		64
#define MAX_EDG_BUF_SIZE	8
#define MAX_DIR_BUF_SIZE	5
#define MAX_VAL_BUF_SIZE	2

unsigned int ngpios;

enum {
		VERSION_OPTION = CHAR_MAX + 1,
};

static void print_usage(char *prg)
{
	fprintf(stderr, "Usage: %s <gpio-number> [Options]\n"
		"Options:\n"
		" -x, --export\t\tExport GPIO pin\n"
		" -d, --direction=DIR\tGPIO set direction (out or in)\n"
		"\t\t\tNote: To ensure glitch free operation use \"high\" or\n"
		"\t\t\t\"low\" to configure as output with that initial value\n"
		" -v, --value=VAL\tGPIO set value (0 or 1)\n"
		" -e, --edge=EDG\t\tGPIO interrupt (rising or falling)\n"
		" -s, --show\t\tDisplay GPIO status\n"
		" -i, --iloop=COUNT\tRun GPIO isr COUNT times\n"
		" -l, --loop=COUNT\tToggle GPIO value COUNT times\n"
		" -u, --unexport\t\tUnexport GPIO pin\n"
		" -h, --help\t\tthis help\n"
		"     --version\t\tprint version information and exit\n"
		"\nExample:\t\t# gpio-sysfs 0 -x\n"
		"\t\t\t# gpio-sysfs 0 -d out -v 1\n\n",
		prg);

	exit(EXIT_FAILURE);
}

static int gpio_fd_open(unsigned int gpio, char *fbuf, int mode)
{
	int fd;
	char buf[MAX_BUF_SIZE];

	if (!strcmp(fbuf, "export") || !strcmp(fbuf, "unexport"))
		snprintf(buf, sizeof(buf), SYSFS_GPIO_DIR "/%s", fbuf);
	else
		snprintf(buf, sizeof(buf), SYSFS_GPIO_DIR "/gpio%d/%s",
							gpio, fbuf);
	fd = open(buf, mode);
	if (fd < 0)
		fprintf(stderr, "failed to open\n", fbuf);

	return fd;
}

static int gpio_fd_close(int fd)
{
	return close(fd);
}

static int gpio_is_valid(unsigned int gpio)
{
	unsigned int gpio_num;
	char buf[MAX_BUF_SIZE];
	char nbuf[MAX_VAL_BUF_SIZE + 1];
	int fd;

	for (gpio_num = 0; gpio_num < 300; gpio_num += 16) {
		snprintf(buf, sizeof(buf), SYSFS_GPIO_DIR "/gpiochip%d/ngpio",
								gpio_num);
		fd = open(buf, O_RDONLY);
		if (fd >= 0) {
			read(fd, nbuf, MAX_VAL_BUF_SIZE + 1);
			ngpios += atoi(nbuf);
		}
	}

	if (gpio < ngpios)
		return 0;
	else
		return -EINVAL;
}

void gpio_export(unsigned int gpio)
{
	int fd, len;
	char buf[MAX_BUF_SIZE];

	fd = gpio_fd_open(gpio, "export", O_WRONLY);
	len = snprintf(buf, sizeof(buf), "%d", gpio);
	write(fd, buf, len);
	gpio_fd_close(fd);
}

void gpio_unexport(unsigned int gpio)
{
	int fd, len;
	char buf[MAX_BUF_SIZE];

	fd = gpio_fd_open(gpio, "unexport", O_WRONLY);
	len = snprintf(buf, sizeof(buf), "%d", gpio);
	write(fd, buf, len);
	gpio_fd_close(fd);
}

void gpio_set_direction(unsigned int gpio, unsigned int direction)
{
	int fd;

	fd = gpio_fd_open(gpio, "direction", O_WRONLY);

	if (direction == GPIO_DIR_HIGH)
		write(fd, "high", MAX_DIR_BUF_SIZE);
	else if (direction == GPIO_DIR_LOW)
		write(fd, "low", MAX_DIR_BUF_SIZE - 1);
	else if (direction == GPIO_DIR_OUT)
		write(fd, "out", MAX_DIR_BUF_SIZE - 1);
	else
		write(fd, "in", MAX_DIR_BUF_SIZE - 2);

	gpio_fd_close(fd);
}

unsigned int gpio_get_direction(unsigned int gpio)
{
	int fd;
	char buf[MAX_DIR_BUF_SIZE];

	fd = gpio_fd_open(gpio, "direction", O_RDONLY);
	read(fd, buf, MAX_DIR_BUF_SIZE);
	gpio_fd_close(fd);

	if (!strncmp(buf, "high", MAX_DIR_BUF_SIZE - 1))
		return GPIO_DIR_HIGH;
	else if (!strncmp(buf, "low", MAX_DIR_BUF_SIZE - 2))
		return GPIO_DIR_LOW;
	else if (!strncmp(buf, "out", MAX_DIR_BUF_SIZE - 2))
		return GPIO_DIR_OUT;

	return GPIO_DIR_IN;
}

void gpio_set_value(unsigned int gpio, unsigned int value)
{
	int fd;

	fd = gpio_fd_open(gpio, "value", O_WRONLY);

	if (value == 0)
		write(fd, "0", MAX_VAL_BUF_SIZE);
	else
		write(fd, "1", MAX_VAL_BUF_SIZE);

	gpio_fd_close(fd);
}

unsigned int gpio_get_value(unsigned int gpio)
{
	int fd;
	char buf[MAX_VAL_BUF_SIZE];

	fd = gpio_fd_open(gpio, "value", O_RDONLY);
	read(fd, buf, MAX_VAL_BUF_SIZE);
	gpio_fd_close(fd);

	return atoi(buf);
}

void gpio_set_edge(unsigned int gpio, unsigned int edge)
{
	int fd;

	fd = gpio_fd_open(gpio, "edge", O_WRONLY);

	if (edge == GPIO_EDG_FALL)
		write(fd, "falling", MAX_EDG_BUF_SIZE);
	else if (edge == GPIO_EDG_RISE)
		write(fd, "rising", MAX_EDG_BUF_SIZE - 1);
	else
		write(fd, "none", MAX_EDG_BUF_SIZE - 3);

	gpio_fd_close(fd);
}

unsigned int gpio_get_edge(unsigned int gpio)
{
	int fd;
	char buf[MAX_EDG_BUF_SIZE];

	fd = gpio_fd_open(gpio, "edge", O_RDONLY);
	read(fd, buf, MAX_EDG_BUF_SIZE);
	gpio_fd_close(fd);

	if (!strncmp(buf, "falling", MAX_EDG_BUF_SIZE - 1))
		return GPIO_EDG_FALL;
	else if (!strncmp(buf, "rising", MAX_EDG_BUF_SIZE - 2))
		return GPIO_EDG_RISE;

	return GPIO_EDG_NONE;
}

int gpio_isr(unsigned int gpio, unsigned int loopcount)
{
	struct pollfd fdset[2];
	int nfds = 2;
	int fd, rc, len;
	char buf[MAX_BUF_SIZE];

	fd = gpio_fd_open(gpio, "value", O_RDONLY | O_NONBLOCK);
	while (loopcount--) {
		memset((void *)fdset, 0, sizeof(fdset));

		fdset[0].fd = STDIN_FILENO;
		fdset[1].fd = fd;
		fdset[0].events = POLLIN;
		fdset[1].events = POLLPRI;

		rc = poll(fdset, nfds, POLL_TIMEOUT);
		if (rc < 0) {
			fprintf(stderr, "poll() failed!\n");
			return -1;
		}

		if (rc == 0)
			fprintf(stdout, ".");

		if (fdset[1].revents & POLLPRI) {
			len = read(fdset[1].fd, buf, MAX_BUF_SIZE);
			fprintf(stdout, "\npoll() GPIO %d interrupt occurred\n",
									gpio);
		}

		if (fdset[0].revents & POLLIN) {
			(void)read(fdset[0].fd, buf, 1);
			fprintf(stdout, "\npoll() stdin read 0x%2.2X\n",
					(unsigned int) buf[0]);
		}
		fflush(stdout);
	}
	fprintf(stdout, "\n");

	gpio_fd_close(fd);
}

static void cmd_show(unsigned int gpio)
{
	unsigned int direction;
	unsigned int value;
	unsigned int edge;

	edge = gpio_get_edge(gpio);
	value = gpio_get_value(gpio);
	direction = gpio_get_direction(gpio);

	fprintf(stdout, "total %d gpios are available\n", ngpios);
	fprintf(stdout, "gpio-%d", gpio);
	if (direction == GPIO_DIR_OUT)
		fprintf(stdout, "\t(sysfs\t\t) out %d", value);
	else
		fprintf(stdout, "\t(sysfs\t\t) in %d", value);

	if (edge == GPIO_EDG_RISE)
		fprintf(stdout, " irq-sysfs edge-rising\n");
	else if (edge == GPIO_EDG_FALL)
		fprintf(stdout, " irq-sysfs edge-falling\n");
	else
		fprintf(stdout, " irq-sysfs edge-none\n");
}

static void cmd_value(unsigned int gpio, unsigned int setvalue)
{
	unsigned int value;

	value = gpio_get_value(gpio);

	/* Set value if the requested value is different from current value */
	if (value != setvalue) {
		gpio_set_value(gpio, setvalue);
		/* Re-check the status, print error if value not changed */
		value = gpio_get_value(gpio);
		if (value != setvalue)
			fprintf(stderr, "gpio%d value is not changed, check"
				" pinmux settings\n", gpio);
		else
			fprintf(stdout, "gpio%d value is changed to %d\n",
				gpio, setvalue);
	}
}

static void cmd_direction(unsigned int gpio, unsigned int setdirection)
{
	unsigned int direction;

	direction = gpio_get_direction(gpio);
	if (direction != setdirection)
		gpio_set_direction(gpio, setdirection);

	direction = gpio_get_direction(gpio);
	if (direction == setdirection) {
		fprintf(stdout, "gpio%d direction changed to", gpio);
		switch (direction) {
		case GPIO_DIR_HIGH:
			fprintf(stdout, " out high\n");
			break;
		case GPIO_DIR_LOW:
			fprintf(stdout, " out low\n");
			break;
		case GPIO_DIR_OUT:
			fprintf(stdout, " out\n");
			break;
		case GPIO_DIR_IN:
		default:
			fprintf(stdout, " in\n");
			break;
		}
	}
}

static void cmd_edge(unsigned int gpio, unsigned int setedge)
{
	unsigned int edge;

	edge = gpio_get_edge(gpio);
	if (edge != setedge)
		gpio_set_edge(gpio, setedge);

	edge = gpio_get_edge(gpio);
	if (edge == setedge) {
		fprintf(stdout, "gpio%d edge trigger changed to", gpio);
		switch (edge) {
		case GPIO_EDG_RISE:
			fprintf(stdout, " rising edge\n");
			break;
		case GPIO_EDG_FALL:
			fprintf(stdout, " falling edge\n");
			break;
		case GPIO_EDG_NONE:
		default:
			fprintf(stdout, " none\n");
			break;
		}
	}
}

static void cmd_interrupt(unsigned int gpio, unsigned int iloop)
{
	unsigned int_flag;

	gpio_set_direction(gpio, GPIO_DIR_IN);
	gpio_isr(gpio, iloop);
}

int main(int argc, char *argv[])
{
	int opt, ret;
	unsigned int gpio;
	unsigned int export = 0;
	unsigned int unexport = 0;
	unsigned int show = 0;
	unsigned int edge = 0;
	unsigned int value = 2;
	unsigned int direction = 0;
	unsigned int loopcount = 0;
	unsigned int int_loopcnt = 0;

	struct option long_options[] = {
		{ "direction",	required_argument,	0, 'd' },
		{ "edge",	required_argument,	0, 'e' },
		{ "help",	no_argument,		0, 'h' },
		{ "int",	required_argument,	0, 'i' },
		{ "loop",	required_argument,	0, 'l' },
		{ "show",	no_argument,		0, 's' },
		{ "unexport",	no_argument,		0, 'u' },
		{ "value",	required_argument,	0, 'v' },
		{ "version",	no_argument,		0, VERSION_OPTION},
		{ "export",	no_argument,		0, 'x' },
		{ 0,		0,			0, 0 },
	};

	if (argc < 2) {
		print_usage(argv[0]);
		exit(0);
	}

	gpio = atoi(argv[1]);
	ret = gpio_is_valid(gpio);
	if (ret < 0) {
		fprintf(stderr, "Error: provide correct gpio number\n");
		exit(EXIT_FAILURE);
	}

	while ((opt = getopt_long(argc, argv, "d:e:i:l:v:hsux",
					long_options, NULL)) != -1) {
		switch (opt) {
		case 'd':
			if (!strcmp(optarg, "high"))
				direction = GPIO_DIR_HIGH;
			else if (!strcmp(optarg, "low"))
				direction = GPIO_DIR_LOW;
			else if (!strcmp(optarg, "out"))
				direction = GPIO_DIR_OUT;
			else
				direction = GPIO_DIR_IN;
			break;

		case 'e':
			if (!strcmp(optarg, "falling"))
				edge = GPIO_EDG_FALL;
			else if (!strcmp(optarg, "rising"))
				edge = GPIO_EDG_RISE;
			else
				edge = GPIO_EDG_NONE;
			break;

		case 'h':
			print_usage(argv[0]);
			exit(0);

		case 'i':
			int_loopcnt = strtoul(optarg, NULL, 0);
			break;

		case 'l':
			loopcount = strtoul(optarg, NULL, 0);
			break;

		case 's':
			show = 1;
			break;

		case 'u':
			unexport = 1;
			break;

		case 'v':
			value = strtoul(optarg, NULL, 0);
			break;

		case 'x':
			export = 1;
			break;

		case VERSION_OPTION:
			fprintf(stdout, "%s %s\n", argv[0], VERSION);
			exit(0);

		default:
			fprintf(stderr, "Unknown option %c\n", opt);
			break;
		}
	}

	if (export)
		gpio_export(gpio);

	if (show)
		cmd_show(gpio);

	if (direction)
		cmd_direction(gpio, direction);

	if ((value == 0) || (value == 1))
		cmd_value(gpio, value);

	if (edge)
		cmd_edge(gpio, edge);

	if (int_loopcnt)
		cmd_interrupt(gpio, int_loopcnt);

	if (loopcount) {
		cmd_direction(gpio, GPIO_DIR_HIGH);
		while (loopcount--) {
			cmd_value(gpio, 0);
			cmd_value(gpio, 1);
		}
	}

	if (unexport)
		gpio_unexport(gpio);

	exit(EXIT_SUCCESS);
}
