/*
 * Ethernet Switch configuration management
 *
 * Copyright (C) 2012 Texas Instruments
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation version 2.
 *
 * This program is distributed "as is" WITHOUT ANY WARRANTY of any
 * kind, whether express or implied; without even the implied warranty
 * of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <stddef.h>
#include <getopt.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <linux/sockios.h>
#include <linux/net_switch_config.h>

#define SWITCH_CONFIG_COMMAND(__var__, __cmd__)			\
	if((__var__) == CONFIG_SWITCH_INVALID) {		\
		(__var__) = (__cmd__);				\
	} else {						\
		printf("Two or more commands Cannot be "	\
			"processed simultaneously\n");		\
		return -1;					\
	}

static char options[] = "?m:u:O:f:x:X:i:F:y:p:t:h::cPTHlIU8AS:Gg:Rn:V:sk:abEj:K:M:N:Y:"
				"z:Z:r:Cq:Q:W:B:L:dewJ:K:D:v";

static struct option long_options[] =
	{
		/* These options set a flag. */
{"add-multi",			required_argument	, 0, 'm'},
{"add-uni",			required_argument	, 0, 'u'},
{"add-oui",			required_argument	, 0, 'O'},
{"find-addr",			required_argument	, 0, 'f'},
{"del-multi",			required_argument	, 0, 'x'},
{"del-uni",			required_argument	, 0, 'X'},
{"add-vlan",			required_argument	, 0, 'i'},
{"find-vlan",			required_argument	, 0, 'F'},
{"del-vlan",			required_argument	, 0, 'y'},
{"port-vlan",			required_argument	, 0, 'p'},
{"timeout",			required_argument	, 0, 't'},
{"dump",			optional_argument	, 0, 'h'},
{"flow-control",		no_argument		, 0, 'c'},
{"prio-map",			no_argument		, 0, 'P'},
{"en-stat",			no_argument		, 0, 'T'},
{"conf-dump",			no_argument		, 0, 'H'},
{"rate-limit",			no_argument		, 0, 'l'},
{"vid-ingress-check",		no_argument		, 0, 'I'},
{"add-unknown-vlan-info",	no_argument		, 0, 'U'},
{"802-1",			no_argument		, 0, '8'},
{"mac-auth",			no_argument		, 0, 'A'},
{"set-port-config",		required_argument	, 0, 'S'},
{"get-port-config",		no_argument		, 0, 'G'},
{"port-state",			required_argument	, 0, 'g'},
{"reset",			no_argument		, 0, 'R'},
{"port",			required_argument	, 0, 'n'},
{"vid",				required_argument	, 0, 'V'},
{"super",			no_argument		, 0, 's'},
{"fwd-state",			required_argument	, 0, 'k'},
{"ageable",			no_argument		, 0, 'a'},
{"blocked",			no_argument		, 0, 'b'},
{"secure",			no_argument		, 0, 'E'},
{"index",			no_argument		, 0, 'j'},
{"vid-untag",			required_argument	, 0, 'K'},
{"reg-multi",			required_argument	, 0, 'M'},
{"unreg-multi",			required_argument	, 0, 'N'},
{"vid-untag-port-mask",		required_argument	, 0, 'Y'},
{"reg-multi-port-mask",		required_argument	, 0, 'z'},
{"unreg-multi-port-mask",	required_argument	, 0, 'Z'},
{"priority",			required_argument	, 0, 'r'},
{"cfi",				no_argument		, 0, 'C'},
{"prio-rx",			required_argument	, 0, 'q'},
{"prio-tx",			required_argument	, 0, 'Q'},
{"prio-swi",			required_argument	, 0, 'W'},
{"addr-type",			required_argument	, 0, 'B'},
{"limit",			required_argument	, 0, 'L'},
{"direction",			no_argument		, 0, 'd'},
{"enable",			no_argument		, 0, 'e'},
{"drop-untagged",		no_argument		, 0, 'w'},
{"unknown-vlan-mem",		required_argument	, 0, 'J'},
{"duplex",			required_argument	, 0, 'D'},
{"version",			no_argument		, 0, 'v'},
{0, 0, 0, 0}
};

static inline int is_zero_ether_addr(const unsigned char *addr)
{
	return !(addr[0] | addr[1] | addr[2] | addr[3] | addr[4] | addr[5]);
}

static inline int is_multicast_ether_addr(const unsigned char *addr)
{
	return 0x01 & addr[0];
}

static inline int is_valid_ether_addr(const unsigned char *addr)
{
	return !is_multicast_ether_addr(addr) && !is_zero_ether_addr(addr);
}

/* String to Hex conversion */
static unsigned char cpmac_str_to_hexnum(unsigned char c)
{
	if(c >= '0' && c <= '9') return c - '0';
	if(c >= 'a' && c <= 'f') return c - 'a' + 10;
	if(c >= 'A' && c <= 'F') return c - 'A' + 10;
	return 0;
}

/* String to ethernet address conversion */
static void config_switch_str_to_ethaddr(char *str, unsigned char *addr)
{
	int i;
	unsigned char num;
	for(i = 0; i < 6; i++) {
		if((*str == '.') || (*str == ':')) str++;
		num = cpmac_str_to_hexnum(*str++) << 4;
		num |= (cpmac_str_to_hexnum(*str++));
		addr[i] = num;
	}
}

void print_help(void)
{
	printf(
		"Switch configuration commands.....\n"
		"All input and output values are in Decimal\n"
		"switch-config -m,--add-multi <address> -n,--port <Portmask> "
			"[-V,--vid <vid>] [-s,--super] "
			"[-k,--fwd-state <value 0-3>]\n"
		"switch-config -u,--add-uni <address> -n,--port <Port number 0-2>"
			" [-V,--vid <vid>] [-a,--ageable] [-b,--blocked] "
			"[-E,--secure]\n"
		"switch-config -O,--add-oui <address>\n"
		"switch-config -f,--find-addr <address> [-V,--vid <vid>]\n"
		"switch-config -x,--del-multi <address> [-V,--vid <vid>] "
			"[-n,--port <Portmask>]\n"
		"switch-config -X,--del-uni <address> [-V,--vid <vid>]\n"
		"switch-config -i,--add-vlan <vlan id> -n,--port <Portmask> "
			"[-K,--vid-untag <Portmask>] "
			"[-M,--reg-multi <Portmask>] "
			"[-N,--unreg-multi <Portmask>]\n"
		"switch-config -F,--find-vlan <vid>\n"
		"switch-config -y,--del-vlan <vid> [-n,--port <Portmask>]\n"
		"switch-config -p,--port-vlan <vid> -n,--port <Port No 0-2> "
			"[-r,--priority <priority 0-7>] [-C,--cfi]\n"
		"switch-config -t,--timeout <timeout (msec)>\n"
		"switch-config -h,--dump [-j,--index <index 0-1023>]\n"
		"switch-config -c,--flow-control -n,--port <Portmask>\n"
		"switch-config -P,--prio-map -n,--port <Portmask> "
			"[-q,--prio-rx <value 0-7>] [-Q,--prio-tx <value 0-7>] "
			"[-W,--prio-swi <value 0-3>]\n"
		"switch-config -T,--en-stat -n,--port <Portmask>\n"
		"switch-config -H,--conf-dump\n"
		"switch-config -l,--rate-limit -n,--port <Portmask> "
			"[-e,--enable (Specify to enable) "
			"-B,--addr-type <broadcast/multicast> "
			"-L,--limit <No of Packet 0-255> "
			"[-d,--direction specify for Tx] ]\n"
		"switch-config -I,--vid-ingress-check -n,--port <Portmask> "
			"[-e,--enable (Specify to enable)] "
			"[-w,--drop-untagged]\n"
		"switch-config -U,--add-unknown-vlan-info "
			"[-Y,--vid-untag-port-mask <0-7>] [-z,--reg-multi-port-mask <0-7>] "
			"[-Z,--unreg-multi-port-mask <0-7>] "
			"[-J,--unknown-vlan-mem <0-7>]\n"
		"switch-config -8,--802-1 [-e,--enable (Specify to enable)]\n"
		"switch-config -A,--mac-auth "
			"[-e,--enable (Specify to enable)]\n"
		"switch-config -S,--set-port-config <0(auto)/10/100/1000> "
			"-n,--port <PortNo> [-D,--duplex <full/half>]\n"
		"switch-config -G,--get-port-config -n,--port <PortNo>\n"
		"switch-config -g,--port-state <disabled/blocked/learn/forward> "
			"-n,--port <PortNo>\n"
		"switch-config -R,--reset\n"
		"\n"
		);
}

int main(int argc, char **argv)
{
	int option_index = 0;
	int c;
	struct net_switch_config cmd_struct;
	unsigned int port_num = 0;
	struct ifreq ifr;
	int sockfd;  /* socket fd we use to manipulate stuff with */

	/* get interface name */
	strncpy(ifr.ifr_name, "eth0", IFNAMSIZ);
	ifr.ifr_data = (char*)&cmd_struct;
	/* Create a channel to the NET kernel. */
	if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		printf("Can't open the socket\n");
		return -1;
	}
	memset(&cmd_struct, 0, sizeof(struct net_switch_config));

	/* parse command line arguments */
	while ((c = getopt_long(argc, argv, options, long_options,
			&option_index)) != -1) {
		switch (c) {
			case '?':
			print_help();
			return 0;

		case 'm':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_ADD_MULTICAST);
			config_switch_str_to_ethaddr(optarg,
				cmd_struct.cmd_data.switchcmd.addr);
		break;

		case 'u':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_ADD_UNICAST);
			config_switch_str_to_ethaddr(optarg,
				cmd_struct.cmd_data.switchcmd.addr);
		break;

		case 'O':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_ADD_OUI);
			config_switch_str_to_ethaddr(optarg,
				cmd_struct.cmd_data.switchcmd.addr);
		break;

		case 'f':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_FIND_ADDR);
			config_switch_str_to_ethaddr(optarg,
				cmd_struct.cmd_data.switchcmd.addr);
		break;

		case 'x':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_DEL_MULTICAST);
			config_switch_str_to_ethaddr(optarg,
				cmd_struct.cmd_data.switchcmd.addr);
		break;

		case 'X':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_DEL_UNICAST);
			config_switch_str_to_ethaddr(optarg,
				cmd_struct.cmd_data.switchcmd.addr);
		break;

		case 'i':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_ADD_VLAN);
			cmd_struct.cmd_data.switchcmd.vid = atoi(optarg);
		break;

		case 'F':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_FIND_VLAN);
			cmd_struct.cmd_data.switchcmd.vid = atoi(optarg);
		break;

		case 'y':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_DEL_VLAN);
			cmd_struct.cmd_data.switchcmd.vid = atoi(optarg);
		break;

		case 'p':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_SET_PORT_VLAN_CONFIG);
			cmd_struct.cmd_data.switchcmd.vid = atoi(optarg);
		break;

		case 't':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_TIMEOUT);
			cmd_struct.cmd_data.switchcmd.ale_timeout= atoi(optarg);
		break;

		case 'h':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_DUMP);
		break;

		case 'c':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_SET_FLOW_CONTROL);
		break;

		case 'P':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_SET_PRIORITY_MAPPING);
		break;

		case 'T':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_PORT_STATISTICS_ENABLE);
		break;

		case 'H':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_CONFIG_DUMP);
		break;

		case 'l':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_RATELIMIT);
		break;

		case 'I':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_VID_INGRESS_CHECK);
		break;

		case 'U':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_ADD_UNKNOWN_VLAN_INFO);
		break;

		case '8':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_802_1);
		break;

		case 'A':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_MACAUTH);
		break;

		case 'S':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_SET_PORT_CONFIG);
			cmd_struct.cmd_data.portcmd.limit =
					atoi(optarg);
		break;

		case 'G':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_GET_PORT_CONFIG);
		break;

		case 'g':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_PORT_STATE);
			if (!strcmp(optarg, "disabled"))
				cmd_struct.cmd_data.portcmd.port_state =
						PORT_STATE_DISABLED;
			else if (!strcmp(optarg, "blocked"))
				cmd_struct.cmd_data.portcmd.port_state =
						PORT_STATE_BLOCKED;
			else if (!strcmp(optarg, "learn"))
				cmd_struct.cmd_data.portcmd.port_state =
						PORT_STATE_LEARN;
			else if (!strcmp(optarg, "forward"))
				cmd_struct.cmd_data.portcmd.port_state =
						PORT_STATE_FORWARD;
			else {
				printf("Invalid Port State\n");
				return -1;
			}
		break;

		case 'R':
			SWITCH_CONFIG_COMMAND(cmd_struct.cmd,
				CONFIG_SWITCH_RESET);
		break;

		case 'n':
			port_num = atoi(optarg);
		break;

		case 'V':
			cmd_struct.cmd_data.switchcmd.vid = atoi(optarg);
		break;

		case 's':
			cmd_struct.cmd_data.switchcmd.flag = 1;
		break;

		case 'k':
			cmd_struct.cmd_data.switchcmd.untag_port =
					atoi(optarg);
		break;

		case 'a':
			cmd_struct.cmd_data.switchcmd.ageable =1;
		break;

		case 'b':
			cmd_struct.cmd_data.switchcmd.blocked = 1;
		break;

		case 'E':
			cmd_struct.cmd_data.switchcmd.secure = 1;
		break;

		case 'j':
			cmd_struct.cmd_data.switchcmd.aledump =
						atoi(optarg);
		break;

		case 'K':
			cmd_struct.cmd_data.switchcmd.untag_port =
						atoi(optarg);
		break;

		case 'M':
			cmd_struct.cmd_data.switchcmd.reg_multi =
						atoi(optarg);
		break;

		case 'N':
			cmd_struct.cmd_data.switchcmd.unreg_multi =
						atoi(optarg);
		break;

		case 'Y':
			cmd_struct.cmd_data.portcmd.port =
					atoi(optarg);
		break;

		case 'z':
			cmd_struct.cmd_data.portcmd.reg_multi_port_mask =
					atoi(optarg);
		break;

		case 'Z':
			cmd_struct.cmd_data.portcmd.unknown_reg_multi_port_mask =
					atoi(optarg);
		break;

		case 'r':
			cmd_struct.cmd_data.switchcmd.prio_port =
					atoi(optarg);
		break;

		case 'C':
			cmd_struct.cmd_data.switchcmd.CFI_port =
					atoi(optarg);
		break;

		case 'q':
			cmd_struct.cmd_data.priocmd.prio_rx =
					atoi(optarg);
		break;

		case 'Q':
			cmd_struct.cmd_data.priocmd.prio_tx =
					atoi(optarg);
		break;

		case 'W':
			cmd_struct.cmd_data.priocmd.prio_switch =
					atoi(optarg);
		break;

		case 'B':
			if(!strcmp(optarg, "broadcast"))
				cmd_struct.cmd_data.portcmd.addr_type =
						ADDR_TYPE_BROADCAST;
			else if(!strcmp(optarg, "multicast"))
				cmd_struct.cmd_data.portcmd.addr_type =
						ADDR_TYPE_MULTICAST;
			else {
				printf("Invalid arguments...\n");
				return -1;
			}
		break;

		case 'L':
			cmd_struct.cmd_data.portcmd.limit =
					atoi(optarg);
		break;

		case 'd':
			cmd_struct.cmd_data.portcmd.direction = 1;
		break;

		case 'e':
			cmd_struct.cmd_data.portcmd.enable = 1;
		break;

		case 'w':
			cmd_struct.cmd_data.portcmd.drop_untagged = 1;
		break;

		case 'J':
			cmd_struct.cmd_data.portcmd.unknown_vlan_member =
					atoi(optarg);
		break;

		case 'D':
			if (!strcmp(optarg, "full"))
				cmd_struct.cmd_data.portcmd.direction = 1;
			else if (!strcmp(optarg, "half"))
				cmd_struct.cmd_data.portcmd.direction = 0;
			else {
				printf("Invalid Options for Duplex\n");
				close(sockfd);
				return -1;
			}
		break;

		default:
			print_help();
			return 0;
		}
	}

	switch (cmd_struct.cmd) {
		case CONFIG_SWITCH_INVALID:
			print_help();
			return -1;
		break;

		case CONFIG_SWITCH_ADD_MULTICAST:
			if ((cmd_struct.cmd_data.switchcmd.untag_port <= 3) &&
				(cmd_struct.cmd_data.switchcmd.vid <= 4095) &&
				(port_num > 0) && (port_num <= 7) &&
				is_multicast_ether_addr(cmd_struct.cmd_data.switchcmd.addr)) {
				cmd_struct.cmd_data.switchcmd.mem_port = port_num;
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("Multicast add failed\n");
					close(sockfd);
					return -1;
				} else
					printf("Multicast address added successfully\n");
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_ADD_UNICAST:
			if ((!(cmd_struct.cmd_data.switchcmd.addr[0] & 0x01)) &&
				(cmd_struct.cmd_data.switchcmd.vid <= 4095) &&
				(port_num <= 2) &&
				is_valid_ether_addr(cmd_struct.cmd_data.switchcmd.addr)) {
				cmd_struct.cmd_data.switchcmd.mem_port = port_num;
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("Multicast add failed\n");
					close(sockfd);
					return -1;
				} else
					printf("Multicast address added successfully\n");
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_ADD_OUI:
			if(!is_zero_ether_addr(cmd_struct.cmd_data.switchcmd.addr)) {
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("OUI address add failed\n");
					close(sockfd);
					return -1;
				} else
					printf("OUI address added successfully\n");
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_FIND_ADDR:
			if ((cmd_struct.cmd_data.switchcmd.vid <= 4095) &&
				!is_zero_ether_addr(cmd_struct.cmd_data.switchcmd.addr)) {
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("Address Not found\n");
					close(sockfd);
					return -1;
				} else
					printf("Address found in index - %d\n",
							cmd_struct.ret_type);
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_DEL_MULTICAST:
			if ((cmd_struct.cmd_data.switchcmd.vid <= 4095) &&
				is_multicast_ether_addr(cmd_struct.cmd_data.switchcmd.addr)) {
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("Multicast Address Not found\n");
					close(sockfd);
					return -1;
				} else
					printf("Multicast Address Deleted successfully\n");
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_DEL_UNICAST:
			if ((cmd_struct.cmd_data.switchcmd.vid <= 4095) &&
				is_valid_ether_addr(cmd_struct.cmd_data.switchcmd.addr)) {
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("Unicast Address Not found\n");
					close(sockfd);
					return -1;
				} else
					printf("Unicast Address Deleted successfully\n");
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_ADD_VLAN:
			if ((cmd_struct.cmd_data.switchcmd.vid <= 4095) &&
				(port_num > 0) && (port_num <= 7)) {
				cmd_struct.cmd_data.switchcmd.mem_port = port_num;
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("VLAN add failed\n");
					close(sockfd);
					return -1;
				} else
					printf("VLAN added successfully\n");
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_FIND_VLAN:
			if (cmd_struct.cmd_data.switchcmd.vid <= 4095) {
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("VLAN Not found\n");
					close(sockfd);
					return -1;
				} else
					printf("VLAN Found\n");
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_DEL_VLAN:
			if ((cmd_struct.cmd_data.switchcmd.vid <= 4095) &&
				(port_num <= 7)) {
				cmd_struct.cmd_data.switchcmd.mem_port = port_num;
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("VLAN Not found\n");
					close(sockfd);
					return -1;
				} else
					printf("VLAN Deleted Successfully\n");
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_SET_PORT_VLAN_CONFIG:
			if ((cmd_struct.cmd_data.switchcmd.vid >= 0) &&
				(cmd_struct.cmd_data.switchcmd.vid <= 4095) &&
				(port_num <= 2) &&
				(cmd_struct.cmd_data.switchcmd.prio_port <= 7)) {
				cmd_struct.cmd_data.switchcmd.port = port_num;
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("Port VLAN Failed\n");
					close(sockfd);
					return -1;
				} else
					printf("Port VLAN Successfully\n");
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_TIMEOUT:
			if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
				printf("Set Timeout Failed\n");
				close(sockfd);
				return -1;
			} else
				printf("Set Timeout Successfully\n");
		break;

		case CONFIG_SWITCH_DUMP:
			if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
				printf("Dump retrieve Failed\n");
				close(sockfd);
				return -1;
			} else {
				printf("Switch Dump\n");
				printf("%s\n", cmd_struct.cmd_data.buf);
			}
		break;

		case CONFIG_SWITCH_SET_FLOW_CONTROL:
			if (port_num <= 7) {
				cmd_struct.cmd_data.portcmd.port = port_num;
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("Set Flow Control failed\n");
					close(sockfd);
					return -1;
				} else
					printf("Set flow control successful\n");
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_SET_PRIORITY_MAPPING:
			if ((port_num <= 2) &&
				(cmd_struct.cmd_data.priocmd.prio_rx <= 7) &&
				(cmd_struct.cmd_data.priocmd.prio_tx <= 7) &&
				(cmd_struct.cmd_data.priocmd.prio_switch<= 3)) {
				cmd_struct.cmd_data.portcmd.port = port_num;
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("Set Priority Mapping Failed\n");
					close(sockfd);
					return -1;
				} else
					printf("Set Priority Mapping successful\n");
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_PORT_STATISTICS_ENABLE:
			if (port_num <= 7) {
				cmd_struct.cmd_data.portcmd.port = port_num;
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("Set Priority Mapping Failed\n");
					close(sockfd);
					return -1;
				} else
					printf("Set Priority Mapping successful\n");
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_CONFIG_DUMP:
			if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
				printf("Switch config Dump retrieve Failed\n");
				close(sockfd);
				return -1;
			} else {
				printf("Switch config Dump\n");
				printf("%s\n", cmd_struct.cmd_data.buf);
			}
		break;

		case CONFIG_SWITCH_RATELIMIT:
			if ((port_num <= 2) &&
				(cmd_struct.cmd_data.portcmd.enable == 0)) {
				cmd_struct.cmd_data.portcmd.port = port_num;
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("Set Rate Limit Failed\n");
					close(sockfd);
					return -1;
				} else
					printf("Set Rate Limit successful\n");
			} else if ((port_num <= 2) &&
				((cmd_struct.cmd_data.portcmd.addr_type == ADDR_TYPE_BROADCAST) ||
				(cmd_struct.cmd_data.portcmd.addr_type == ADDR_TYPE_MULTICAST)) &&
				(cmd_struct.cmd_data.portcmd.limit <= 255)) {
				cmd_struct.cmd_data.portcmd.port = port_num;
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("Set Rate Limit Failed\n");
					close(sockfd);
					return -1;
				} else
					printf("Set Rate Limit successful\n");
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_VID_INGRESS_CHECK:
			if (port_num <= 2) {
				cmd_struct.cmd_data.portcmd.port = port_num;
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("VID Ingress Check Failed\n");
					close(sockfd);
					return -1;
				} else
					printf("VID Ingress Check successful\n");
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_ADD_UNKNOWN_VLAN_INFO:
			if ((cmd_struct.cmd_data.portcmd.reg_multi_port_mask <= 7) &&
				(cmd_struct.cmd_data.portcmd.unknown_reg_multi_port_mask <= 7) &&
				(cmd_struct.cmd_data.portcmd.unknown_vlan_member <= 7)) {
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("Unknown VLAN Info Failed\n");
					close(sockfd);
					return -1;
				} else
					printf("Unknown VLAN Info successful\n");
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_802_1:
			if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
				printf("802.1x Support Failed\n");
				close(sockfd);
				return -1;
			} else
				printf("802.1x Support Sucessful\n");
		break;

		case CONFIG_SWITCH_MACAUTH:
			if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
				printf("MAC Auth Failed\n");
				close(sockfd);
				return -1;
			} else
				printf("MAC Auth Sucessful\n");
		break;

		case CONFIG_SWITCH_SET_PORT_CONFIG:
			if (((port_num == 1) || (port_num == 2)) &&
				((cmd_struct.cmd_data.portcmd.limit == 0) ||
				(cmd_struct.cmd_data.portcmd.limit == 10) ||
				(cmd_struct.cmd_data.portcmd.limit == 100) ||
				(cmd_struct.cmd_data.portcmd.limit == 1000))) {
				cmd_struct.cmd_data.portcmd.port = port_num;
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("Set Port Config Failed\n");
					close(sockfd);
					return -1;
				} else
					printf("Set Port Config successful\n");
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_GET_PORT_CONFIG:
			if ((port_num == 1) || (port_num == 2)) {
				cmd_struct.cmd_data.portcmd.port = port_num;
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("Get Port Config Failed\n");
					close(sockfd);
					return -1;
				} else {
					printf("Get Port Config successful\n");
					printf("Speed : %d\n", cmd_struct.cmd_data.portcmd.limit);
					if(cmd_struct.cmd_data.portcmd.direction)
						printf("Duplexity FULL\n");
					else
						printf("Duplexity FULL\n");
				}
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_PORT_STATE:
			if (port_num <= 2) {
				cmd_struct.cmd_data.portcmd.port = port_num;
				if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
					printf("Set Port State Failed\n");
					close(sockfd);
					return -1;
				} else
					printf("Set Port State successful\n");
			} else {
				printf("Invalid Arguments\n");
				return -1;
			}
		break;

		case CONFIG_SWITCH_RESET:
			if (ioctl(sockfd, SIOCSWITCHCONFIG, &ifr) < 0) {
				printf("Reset Failed\n");
				close(sockfd);
				return -1;
			} else
				printf("Reset successful\n");
		break;
	}

	close(sockfd);
	return 0;
}
